/* This is the widget for drawing of input points together with their 2D convex hull (CH).
 * The user can select input point distribution, number of input points and the CH algorithm. 
 * 
 * concept and design: Petr Felkel - felkepet@fel.cvut.cz
 *
 * programming: Petr Felkel - felkepet@fel.cvut.cz
 *				Lukas Novosad - novosluk@fel.cvut.cz
 *
 * date: 21.10.2010
 *
 */

#include "drawareawidget.h"
#include <vector>

// CGAL includes
#include <CGAL/ch_jarvis.h>
#include <CGAL/ch_eddy.h>

// your header file for Graham and Jarvis algorithms
#include "ch_graham_jarvis.h"


// the global std vector for the generated points
//std::vector<Point_2> points;

DrawAreaWidget::DrawAreaWidget(QWidget *parent)
	: QWidget(parent), window_center(0,0)
{
	ui.setupUi(this); 

	// intialize the point generator, CH algorithm, and the number of points
	mPoint_generator = IN_DISC;
	mChAlgorithm = AKL;
	mNumberOfPoints = 10;

	// initialize the iso-rectangle and segment corners
	P1 = Point_2(-220,-120);
	P2 = Point_2( 220, 160);

	// generate the initial points
	generatePoints( mNumberOfPoints, mPoints );

	// compute the convex hull according to the selected CH algorithm
	generateConvexHull( mPoints, mConvexHull );
}

DrawAreaWidget::~DrawAreaWidget()
{
	
}



// slot for switching between the point generators
void DrawAreaWidget::setPointGenerator(PointGenerator _generator)
{
	mPoint_generator = _generator;
	QString message = QString("Points are generated %1").arg(getPointGeneratorString());
    emit statusBarMessageChanged(QString(message).append(QString(" ---- measurement started")));  // update the status bar


	generatePoints( mNumberOfPoints, mPoints );


	// erase the previous Convex Hull
	if( !mConvexHull.empty() )  
		mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );

	repaint(); // show the points before the computation of CH

	timer.Reset();
	int repeatTimes = (mNumberOfPoints < 10000) ? 10000 / mNumberOfPoints  : 1;
	for( int i = 0; i < repeatTimes; i++ )
	{
		// erase the previous Convex Hull
		if( !mConvexHull.empty() )  
			mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );

		// compute the convex hull according to the selected CH algorithm
		generateConvexHull( mPoints, mConvexHull );
	}

	repaint(); // show the points and CH

	// update the status bar
	emit statusBarMessageChanged(QString(message).append(getTimeMeasurementString()));
}



// slot for setting the number of points to be generated
void DrawAreaWidget::setNumberOfPoints(long value)
{ 
	mNumberOfPoints = value;

	QString message = QString("Number of points changed to %1").arg(mNumberOfPoints);  // text for status bar
    emit statusBarMessageChanged(QString(message).append(QString(" ---- measurement started")));  // update the status bar

	generatePoints( mNumberOfPoints, mPoints );

	// erase the previous Convex Hull
	if( !mConvexHull.empty() )  
		mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );
	
	repaint(); // show the points before the computation of CH

	timer.Reset();
	int repeatTimes = (mNumberOfPoints < 10000) ? 10000 / mNumberOfPoints  : 1;
	for( int i = 0; i < repeatTimes; i++ )
	{
		// erase the previous Convex Hull
		if( !mConvexHull.empty() )  
			mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );

		// compute the convex hull according to the selected CH algorithm
		generateConvexHull( mPoints, mConvexHull );
	}

	repaint(); // show the points and CH

	// update the status bar
	emit statusBarMessageChanged(QString(message).append(getTimeMeasurementString()));
}

// slot for selecting the CH algorithm
void DrawAreaWidget::setChAlgorithm(ChAlgorithm _chAlg)
{
	mChAlgorithm = _chAlg;

	QString message = QString("CH algorithm changed to %1").arg(getChAlgorithmString());
    emit statusBarMessageChanged(QString(message).append(QString(" ---- measurement started")));  // update the status bar


	// erase the previous Convex Hull
	if( !mConvexHull.empty() )  
		mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );

	repaint(); // show the points before the computation of CH

	timer.Reset();
	int repeatTimes = (mNumberOfPoints < 10000) ? 10000 / mNumberOfPoints  : 1;
	for( int i = 0; i < repeatTimes; i++ )
	{
		// erase the previous Convex Hull
		if( !mConvexHull.empty() )  
			mConvexHull.erase( mConvexHull.begin(), mConvexHull.end() );

		// compute the convex hull according to the selected CH algorithm
		generateConvexHull( mPoints, mConvexHull );
	}

	repaint(); // show the points and CH

	// update the status bar
	emit statusBarMessageChanged(QString(message).append(getTimeMeasurementString()));
}


/*
 * This method creates a measurement string from the timings returned by the timer.
 */
QString DrawAreaWidget::getTimeMeasurementString()
{
	QString measurementString;
	double time = timer.RealTime();
	//measurementString.append(QString(" ---- time needed to compute the convex hull is: %1 [ms] (user time)").arg(timer.UserTime() * 1000));
	//measurementString.append(QString(" ---- time needed to compute the convex hull is: %1 [ms] (real time)").arg(timer.RealTime() * 1000));
	//measurementString.append(QString(" ---- time needed to compute the convex hull is: %1 [ms] (system time)").arg(timer.SystemTime() * 1000));
	measurementString.append(QString(" ---- time needed to compute the convex hull is: %1 [ms] (user time) / %2 [ms] (real time)").arg(timer.UserTime()*1000).arg(timer.RealTime() * 1000));
    return measurementString;
}


/*
 * This method is used to format the output on the status bar.
 */
QString DrawAreaWidget::getPointGeneratorString()
{
	switch( mPoint_generator )
	{
	case IN_DISC:
		return QString("on a disc");
		break;
	case ON_CIRCLE:
		return QString("on a circle");
		break;
	case IN_SQUARE:
		return QString("in a square");
		break;
	case ON_SQUARE:
		return QString("on a square");
		break;
	case IN_RECTANGLE:
		return QString("in a rectangle");
		break;
	case ON_SEGMENT:
		return QString("on a segment");
		break;
	default:
		return QString();
		break;
	}
}


/*
 * This method is used to format the output on the status bar.
 */
QString DrawAreaWidget::getChAlgorithmString()
{
	switch( mChAlgorithm ) 
	{
	case AKL:
			return QString("Akl");
			break;
	case BYKAT:
			return QString("Bykat");
			break;
	case EDDY:
			return QString("Eddy");
			break;
	case GRAHAM_ANDREW:
			return QString("Graham_andrew");
			break;
	case JARVIS:
			return QString("Jarvis");
			break;
	case MY_GRAHAM:
			return QString("My_Graham");
			break;
	case MY_JARVIS:
			return QString("My_Jarvis");
			break;
	default:
		return QString();
		break;
	}
}


/*
	This method implements the resize event.
	It overrides the virtual definition from the QWidget class.
*/
void DrawAreaWidget::resizeEvent( QResizeEvent * event )
{
	window_center.setX(event->size().width()  / 2);  
	window_center.setY(event->size().height() / 2);
}

/*
	This method generates the points based on the number of points and the point distribution selected by the user.
*/
void DrawAreaWidget::generatePoints( long n, std::vector<Point_2> &points )
{	
	int radius = DISC_DIAMETER;
	
	CGAL::Random_points_in_disc_2< Point_2, PointCreator> genInDisc(radius);				// generated range -+radius
	CGAL::Random_points_on_circle_2< Point_2, PointCreator> genOnCircle(radius);			// generated range -+radius
	CGAL::Random_points_in_square_2< Point_2, PointCreator> genInSquare(radius);			// generated range -+radius
	CGAL::Random_points_on_square_2< Point_2, PointCreator> genOnSquare(radius);			// generated range -+radius
	CGAL::Random_points_in_iso_rectangle_2< Point_2, PointCreator> genInRectangle(P1, P2 ); // P1 = Point_2(-220,-120);
	CGAL::Random_points_on_segment_2< Point_2, PointCreator> genOnSegment(P1, P2 );			// P2 = Point_2( 220, 160);

	points.erase( points.begin(), points.end() );
	points.reserve( n );

	switch( mPoint_generator ) // generate the input points
	{
		case IN_DISC: 		CGAL::copy_n( genInDisc, n, std::back_inserter(points) );		break;
		case ON_CIRCLE:		CGAL::copy_n( genOnCircle, n, std::back_inserter(points) );		break;
		case IN_SQUARE:		CGAL::copy_n( genInSquare, n, std::back_inserter(points) );		break;
		case ON_SQUARE:		CGAL::copy_n( genOnSquare, n, std::back_inserter(points) );		break;
		case IN_RECTANGLE:	CGAL::copy_n( genInRectangle, n, std::back_inserter(points) );	break;
		case ON_SEGMENT:	CGAL::copy_n( genOnSegment, n, std::back_inserter(points) );	break;
	}

}


/*
	This method computes the convex hull.
*/
void DrawAreaWidget::generateConvexHull( std::vector<Point_2> &points, std::vector<Point_2> &ch)
{
	//timer.Reset();
	if( !ch.empty() )  
		ch.erase( ch.begin(), ch.end() );

	// switch according to the user's selection and call the appropriate convex hull algorithm
	switch( mChAlgorithm ) 
	{
	case AKL:
		timer.Start();
		CGAL::ch_akl_toussaint( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case BYKAT:
		timer.Start();
		CGAL::ch_bykat( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case EDDY:
		timer.Start();
		CGAL::ch_eddy( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case GRAHAM_ANDREW:
		timer.Start();
		CGAL::ch_graham_andrew( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case JARVIS:
		timer.Start();
		CGAL::ch_jarvis( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case MY_GRAHAM:   // TODO VGE: CALL YOUR ALGORITHM HERE
		timer.Start();
		CGAL::convex_hull_points_2( points.begin(), points.end(), std::back_inserter(ch) );
		// ch_my_graham( points.begin(), points.end(), std::back_inserter(ch) ); // implement this procedure
		timer.Stop();
		break;
	case MY_JARVIS:   // TODO VGE: CALL YOUR ALGORITHM HERE
		timer.Start();
		CGAL::convex_hull_points_2( points.begin(), points.end(), std::back_inserter(ch) );
		// ch_my_jarvis( points.begin(), points.end(), std::back_inserter(ch) ); // implement this procedure
		timer.Stop();
		break;
	}
}


/*
	This method draws all the generated points.
*/
void DrawAreaWidget::drawPoints( QPainter &painter, std::vector<Point_2> &points )
{
	// draw all the points in blue
	for (unsigned int i = 0; i < points.size() ; i++)
		painter.fillRect( points[i].x() - POINT_HALFSIZE + window_center.x(), points[i].y()-POINT_HALFSIZE + window_center.y(), POINT_SIZE, POINT_SIZE, Qt::blue);
}


/*
	This method draws the convex hull polygon.
*/
void DrawAreaWidget::drawPolygon( QPainter &painter, std::vector<Point_2> &points )
{
	// set the pen to green for the convex hull
	// remember the painter width
	int oldWidth = painter.pen().width();
	QPen pen(Qt::green);
	pen.setWidth(2);
	painter.setPen(pen);
	
	// create std iterators
	std::vector<Point_2>::iterator it;
	std::vector<Point_2>::iterator it_prev;

	// iterate through all the points of the convex hull and draw the line segments
	it = points.begin();
	if( it != points.end() )
	{
		it_prev = it;
		it++;
		while( it != points.end() )
		{
			painter.drawLine( it->x() + window_center.x(), it->y() + window_center.y(), 
							  it_prev->x() + window_center.x(), it_prev->y() + window_center.y() );
			it_prev = it++;
		}
		painter.drawLine( points.begin()->x() + window_center.x(), points.begin()->y() + window_center.y(), 
						  it_prev->x() + window_center.x(), it_prev->y() + window_center.y() );

	}

	// set the old pen width of the painter
	pen.setWidth(oldWidth);

	// draw the points of the convex hull in red
	for( it = points.begin(); it != points.end() ; it++ )
	{
		painter.fillRect( (*it).x() - 2*POINT_HALFSIZE + window_center.x(), (*it).y()-2*POINT_HALFSIZE + window_center.y(), 2*POINT_SIZE, 2*POINT_SIZE, Qt::red);
	}

}


/*
	This method implements the painting.
	It overrides the virtual definition from the QWidget class.
*/
void DrawAreaWidget::paintEvent(QPaintEvent * paintEvent)
{
	int radius = DISC_DIAMETER;
	
	QPainter painter(this);
	
	// draw the delimiter for the area of points
	painter.setPen(Qt::black);
	switch( mPoint_generator )
	{
		case IN_DISC: 	painter.drawEllipse( window_center, radius, radius ); break;
		case ON_CIRCLE:	painter.drawEllipse( window_center, radius, radius ); break;
		case IN_SQUARE:	painter.drawRect( window_center.x()-radius, window_center.y()-radius, 2*radius, 2*radius );	break;
		case ON_SQUARE:	painter.drawRect( window_center.x()-radius, window_center.y()-radius, 2*radius, 2*radius );	break;
		case IN_RECTANGLE:
			rect.setLeft(window_center.x()+P1.x());
			rect.setBottom(window_center.y()+P1.y());
			rect.setRight(window_center.x()+P2.x());
			rect.setTop(window_center.y()+P2.y());
			painter.drawRect(rect);
			break;
		case ON_SEGMENT: break;
	}

	// draw all the points
	drawPoints( painter, mPoints );

	// draw the convex hull polygon and its points
	drawPolygon( painter, mConvexHull );

}