/* This is the widget for drawing of input points together with their 2D convex hull (CH).
 * The user can select input point distribution, number of input points and the CH algorithm. 
 * 
 * concept and design: Petr Felkel - felkepet@fel.cvut.cz
 *
 * programming: Petr Felkel - felkepet@fel.cvut.cz
 *				Lukas Novosad - novosluk@fel.cvut.cz
 *
 * date: 21.10.2010
 *
 */

#ifndef DRAWAREAWIDGET_H
#define DRAWAREAWIDGET_H

// Qt includes
#include <QPoint>
#include <QWidget>
#include <QPainter>
#include <QPaintEvent>
#include "ui_drawareawidget.h"
#include <QStatusBar>

// CGAL includes
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/point_generators_2.h>
#include <CGAL/algorithm.h>
#include <CGAL/convex_hull_traits_2.h>
#include <CGAL/convex_hull_2.h>

#include "timer.h"


// define how precise the computations should be
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Convex_hull_traits_2<K> Traits;

// define the point creator
typedef K::Point_2 Point_2;
typedef CGAL::Creator_uniform_2<double, Point_2> PointCreator;

// define the types of point genertators and convex hull algorithms
enum PointGenerator { IN_DISC, ON_CIRCLE, IN_SQUARE, ON_SQUARE, IN_RECTANGLE, ON_SEGMENT };
enum ChAlgorithm { AKL, BYKAT, EDDY, GRAHAM_ANDREW, JARVIS, MY_GRAHAM, MY_JARVIS };

// define the constants used for drawing
const int POINT_HALFSIZE = 1;
const int POINT_SIZE = 3 * POINT_HALFSIZE;
const int DISC_DIAMETER = 220;

class DrawAreaWidget : public QWidget
{
	Q_OBJECT

public:
	DrawAreaWidget(QWidget *parent = 0);
	~DrawAreaWidget();

protected:
	void paintEvent(QPaintEvent *);
	void resizeEvent( QResizeEvent * event );

private:
	Ui::DrawAreaWidgetClass ui;

	QPoint window_center;	// the center of the draw widget
	QRect rect;				// used to delimit the point area
	Point_2 P1, P2;			// iso-rectangle and segment corners 

	std::vector<Point_2> mPoints;     // the std vector for the generated points
	std::vector<Point_2> mConvexHull; // std vector which will store our convex hull points

	PointGenerator mPoint_generator;	// the generator type for the points
	ChAlgorithm mChAlgorithm;			// the CH algorithm
	long mNumberOfPoints;				// the number of points

	CTimer timer; // the timer for measurements

	// return the string for the currently selected point type
	QString getPointGeneratorString();
	// return the string for the currently selected CH algorithm
	QString getChAlgorithmString();
	// return the string for the current time measurement
	QString getTimeMeasurementString();

public:
	
	// generate all the points and store them in the passed std vector
	void generatePoints( long n, std::vector<Point_2> &points );
	// compute the convex hull based on the points passed in the std vector points
	void generateConvexHull( std::vector<Point_2> &mPoints, std::vector<Point_2> &ch);
	// draw the points passed in the std vector
	void drawPoints( QPainter &painter, std::vector<Point_2> &points );
	// draw a polygon based on the points passed in the std vector
	void drawPolygon( QPainter &painter, std::vector<Point_2> &points );

public slots:

	// slot to set the point generator type
	void setPointGenerator(PointGenerator _generator);
	// slot to set the CH algorithm to be used
    void setChAlgorithm(ChAlgorithm _chAlg);
	// slot to set the number of points to be generated
	void setNumberOfPoints(long value);

signals:

	// signal to update the status bar in the main window
	void statusBarMessageChanged(const QString &message);

};

#endif // DRAWAREAWIDGET_H
